package scenic3;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import net.arnx.jsonic.JSON;

import org.slim3.controller.Controller;
import org.slim3.controller.Navigation;
import org.slim3.controller.validator.Errors;

import scenic3.util.Files;
import scenic3.util.Stream;

/**
 * Senic3でPageを作成するときに継承するPageクラス。
 * 
 * @author shuji.w6e
 * @since 0.1.0
 */
public abstract class ScenicPage {

    /** ServletContext */
    protected ServletContext servletContext;
    /** HttpServletRequest */
    protected HttpServletRequest request;
    /** HttpServletResponse */
    protected HttpServletResponse response;
    /** base path of request */
    protected String basePath;
    /** Errors */
    protected Errors errors;
    /** original controller */
    ScenicController controller;

    void setController(ScenicController controller) {
        this.controller = controller;
    }
    
    /**
     * Call when page setup.
     * @Deprecated use setUp method.
     */
    @Deprecated
    protected void setupPage() {
    }

    /**
     * Behave Slim3 Controller's same method.
     * @since 0.4.0
     */
    protected Navigation setUp() {
        return null;
    }

    /**
     * Behave Slim3 Controller's same method.
     * @since 0.4.0
     */
    protected void tearDown() {
    }

    /**
     * Delegate Slim3 Controller's same method.
     * @param fileName
     * @param data
     * @since 0.4.0
     */
    protected void show(String fileName, byte[] data) {
        controller.show(fileName, data);
    }

    /**
     * Delegate Slim3 Controller's same method.
     * @param fileName
     * @param in
     * @since 0.4.0
     */
    protected void show(String fileName, InputStream in) {
        controller.show(fileName, in);
    }
    
    /**
     * Delegate Slim3 Controller's same method.
     * @param fileName
     * @param data
     * @since 0.4.1
     */
    protected void download(String fileName, byte[] data) {
        controller.download(fileName, data);
    }

    /**
     * Delegate Slim3 Controller's same method.
     * @param fileName
     * @param in
     * @since 0.4.1
     */
    protected void download(String fileName, InputStream in) {
        controller.download(fileName, in);
    }
    
    public Navigation handleError(Throwable error) throws Throwable {
        throw error;
    }

    /**
     * delegate slim3's isGet method
     * @return
     * @see Controller#isGet
     */
    protected boolean isGet() {
        assert controller != null;
        return controller.isGet();
    }

    /**
     * delegate slim3's isPost method
     * @return
     * @see Controller#isPost
     */
    protected boolean isPost() {
        assert controller != null;
        return controller.isPost();
    }
    
    
    /**
     * delegate slim3's isPut method
     * @return
     * @see Controller#isPut
     */
    protected boolean isPut() {
        assert controller != null;
        return controller.isPut();
    }

    /**
     * delegate slim3's isDelete method
     * @return
     * @see Controller#isDelete
     */
    protected boolean isDelete() {
        assert controller != null;
        return controller.isDelete();
    }
    
    /**
     * delegate slim3's requestScope method
     * @param name
     * @param value
     * @throws NullPointerException
     * @see Controller#requestScope
     */
    protected void requestScope(CharSequence name, Object value) throws NullPointerException {
        controller.requestScope(name, value);
    }

    /**
     * delegate slim3's requestScope method
     * @param <T>
     * @param name
     * @return
     * @throws NullPointerException
     * @see Controller#requestScope
     */
    protected <T> T requestScope(CharSequence name) throws NullPointerException {
        return controller.<T> requestScope(name);
    }

    /**
     * delegate slim3's removeRequestScope method
     * @param <T>
     * @param name
     * @return
     * @throws NullPointerException
     * @see Controller#removeRequestScope
     */
    protected <T> T removeRequestScope(CharSequence name) throws NullPointerException {
        return controller.<T> removeRequestScope(name);
    }

    /**
     * delegate slim3's paramValues method
     * @param <T>
     * @param name
     * @return
     * @throws NullPointerException
     * @see Controller#paramValues
     */
    protected String[] paramValues(CharSequence name) throws NullPointerException {
        return controller.paramValues(name);
    }

    /**
     * delegate slim3's param method
     * @param <T>
     * @param name
     * @return
     * @throws NullPointerException
     * @see Controller#param
     */
    protected String param(CharSequence name) throws NullPointerException {
        return controller.param(name);
    }
    
    /**
     * Delegate Slim3 Controller's same method.
     * @param name
     * @param value
     * @throws NullPointerException
     */
    protected void sessionScope(CharSequence name, Object value) throws NullPointerException {
        controller.sessionScope(name, value);
    }
    
    /**
     * Delegate Slim3 Controller's same method.
     * @param <T>
     * @param name
     * @return
     * @throws NullPointerException
     */
    protected <T> T sessionScope(CharSequence name) throws NullPointerException {
        return controller.<T> sessionScope(name);
    }
    
    /**
     * Delegate Slim3 Controller's same method.
     * @param <T>
     * @param name
     * @return
     * @throws NullPointerException
     */
    protected <T> T removeSessionScope(CharSequence name) throws NullPointerException {
        return controller.<T> removeSessionScope(name);
    }
    
    /**
     * Delegate Slim3 Controller's same method.
     * @param path
     * @return
     */
    protected Navigation forward(String path) {
        assert controller != null;
        return controller.forward(path);
    }
    
    /**
     * Delegate Slim3 Controller's same method.
     * @param path
     * @return
     */
    protected Navigation redirect(String path) {
        assert controller != null;
        return controller.redirect(path);
    }

    protected Navigation response(String text) throws IOException {
        assert controller != null;
        String charset = request.getCharacterEncoding();
        response.setContentType("text/plain; charset=" + charset);
        response.setContentLength(text.getBytes(charset).length);
        response.getWriter().write(text);
        response.flushBuffer();
        return null;
    }
    
    /**
     * send error
     * @param statusCode
     * @return
     * @throws IOException
     * @since 0.4.0
     */
    protected Navigation sendError(int statusCode) throws IOException {
        assert response != null;
        response.sendError(statusCode);
        return null;
    }
    
    /**
     * send error and error message.
     * @param statusCode
     * @param msg
     * @return
     * @throws IOException
     * @since 0.4.0
     */
    protected Navigation sendError(int statusCode, String msg) throws IOException {
        assert response != null;
        response.sendError(statusCode, msg);
        return null;
    }
    
    /**
     * send error with 404 status code.
     * @param msg
     * @return
     * @throws IOException
     */
    protected Navigation notFound(String msg) throws IOException {
        return this.sendError(HttpServletResponse.SC_NOT_FOUND, msg);
    }
    
    protected Navigation responseJson(File file) throws IOException {
        response.setContentType("text/javascript; charset=utf-8");
        String text = Files.readAsString(file, "UTF-8");
        response.getWriter().write(text);
        response.flushBuffer();
        return null;
    }

    protected Navigation responseJson(Object obj) throws IOException {
        response.setContentType("text/javascript; charset=utf-8");
        JSON.encode(obj, response.getOutputStream(), false);
        response.flushBuffer();
        return null;
    }

    protected Navigation responseDirect(InputStream is, String contentType) throws IOException {
        response.setContentType(contentType);
        Stream.copy(is, response.getOutputStream());
        return null;
    }

    protected Navigation responseDirect(File file, String contentType) throws IOException {
        response.setContentType(contentType);
        Stream.copy(Files.newFileInputStream(file), response.getOutputStream());
        return null;
    }

    protected Navigation responseDirect(byte[] bytes, String contentType) throws IOException {
        response.setContentType(contentType);
        response.setContentLength(bytes.length);
        response.getOutputStream().write(bytes);
        response.flushBuffer();
        return null;
    }

}
